package com.lnb.multidatasourcedemo.curator;

import com.google.common.collect.Maps;
import com.lnb.multidatasourcedemo.constant.Constant;
import com.lnb.multidatasourcedemo.datasource.DynamicDataSource;
import com.lnb.multidatasourcedemo.listener.AppListener;
import com.lnb.multidatasourcedemo.listener.DbconfigListener;
import com.lnb.multidatasourcedemo.listener.RoutingKeyListener;
import com.lnb.multidatasourcedemo.listener.ZKConnectionStateListener;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.curator.framework.CuratorFramework;
import org.apache.curator.framework.CuratorFrameworkFactory;
import org.apache.curator.framework.recipes.cache.PathChildrenCache;
import org.apache.curator.retry.RetryUntilElapsed;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * zookeeper客户端配置
 * created by liaoningbo on 2021/11/24
 */
@Component
@Slf4j
public class CuratorClient {

    @Autowired
    public ZkProperties zkProperties;

    private CuratorFramework curator;

    // 指定连接超时和session超时，时间为100秒
    private final int CONNECT_TIMEOUT = 100000;
    private final int RETRY_TIME = Integer.MAX_VALUE;
    // 指定重连间隔为6分钟
    private final int RETRY_INTERVAL = 360000;
    // 监听器列表
    public Map<String, PathChildrenCache> listenerList = Maps.newConcurrentMap();

    @PostConstruct
    public void initCuratorClient() {
        // 基础参数校验
        if (!zkProperties.validateProperties())
            throw new IllegalStateException("Properties config error!");
        String zkServers = zkProperties.getZkServer();
        if (curator == null) {
            curator = newCurator(zkServers);
        }
    }

    private CuratorFramework newCurator(String zkServers) {
        CuratorFramework client = CuratorFrameworkFactory.builder().connectString(zkServers)
                // 指定最大超时时间和重连时间间隔,直到成功为止
                .retryPolicy(new RetryUntilElapsed(RETRY_TIME, RETRY_INTERVAL)).connectionTimeoutMs(CONNECT_TIMEOUT)
                .sessionTimeoutMs(CONNECT_TIMEOUT).build();
        client.start();
        return client;
    }

    /**
     * 注册/dynamicDataSource/dbconfig这一级监听器，用于监听App的增加、删除
     * @param dynamicDataSource
     */
    public void registerAppListenter(DynamicDataSource dynamicDataSource) throws Exception {
        if (!listenerList.containsKey(Constant.DBCONFIG_ROOT)) {
            log.info("Register app Listener start");
            //创建监听
            PathChildrenCache childrenCache = new PathChildrenCache(curator, Constant.DBCONFIG_ROOT, false);
            AppListener appListener = new AppListener(dynamicDataSource, this);
            childrenCache.getListenable().addListener(appListener);
            childrenCache.start(PathChildrenCache.StartMode.BUILD_INITIAL_CACHE);
            //添加到监听列表
            listenerList.put(Constant.DBCONFIG_ROOT, childrenCache);
            log.info("Register app Listener finished successfully");
        } else
            log.info("The listener already exist! path:[" + Constant.DBCONFIG_ROOT + "].");
    }

    /**
     * 注册/dynamicDataSource/dbconfig/xxx(appname)这一级监听器，用于监听数据库实例的增加、删除
     */
    public void registerDbconfigListenter(DynamicDataSource dynamicDataSource, String app) throws Exception {
        Set<String> apps = new HashSet<>();
        if (StringUtils.isBlank(app)) {
            apps.addAll(zkProperties.getAppName());
        } else {
            apps.add(app);
        }
        for (String appName : apps) {
            if (StringUtils.isNotBlank(appName)) {
                String appDB = Constant.getAppDB(appName);
                if (!listenerList.containsKey(appDB)) {
                    log.info("Registering app DB:{} Listener", appDB);
                    //创建监听
                    PathChildrenCache childrenCache = new PathChildrenCache(curator, appDB, false);
                    DbconfigListener dbconfigListener = new DbconfigListener(dynamicDataSource, this);
                    childrenCache.getListenable().addListener(dbconfigListener);
                    childrenCache.start(PathChildrenCache.StartMode.BUILD_INITIAL_CACHE);
                    //添加到监听列表
                    listenerList.put(appDB, childrenCache);
                    log.info("Register app DB:{} Listener finished successfully", appDB);
                } else
                    log.info("The listener already exist! path:[" + appDB + "].");
            }
        }
    }

    /**
     * 注册/dynamicDataSource/dbconfig/xxx(appname)/xxx这一级监听器，用于监听数据库routingKey的增加、删除
     */
    public void registerRoutingKeyListenter(DynamicDataSource dynamicDataSource, String path, String app)
            throws Exception {
        Set<String> apps = new HashSet<>();
        if (StringUtils.isBlank(app)) {
            apps.addAll(zkProperties.getAppName());
        } else {
            apps.add(app);
        }
        for (String appName : apps) {
            if (StringUtils.isNotBlank(appName)) {
                path = Constant.getAppDB(appName) + Constant.PATH_SPLIT + path;
                if (!listenerList.containsKey(path)) {
                    log.info("Registering routingKey Listener path:{}", path);
                    //创建监听
                    PathChildrenCache childrenCache = new PathChildrenCache(curator, path, false);
                    childrenCache.start(PathChildrenCache.StartMode.POST_INITIALIZED_EVENT);
                    RoutingKeyListener routingKeyListener = new RoutingKeyListener(dynamicDataSource, this);
                    childrenCache.getListenable().addListener(routingKeyListener);
                    //添加到监听列表
                    listenerList.putIfAbsent(path, childrenCache);
                    log.info("Register routingKey Listener finished successfully path:{}", path);
                } else
                    log.info("The listener already exist! path:[" + path + "].");
            }
        }
    }

    /**
     * 监听与ZK的连接状态
     */
    public void registerZKstateListenter(DynamicDataSource dynamicDataSource) throws Exception {
        curator.getConnectionStateListenable()
                .addListener(new ZKConnectionStateListener(dynamicDataSource, this, zkProperties));
        log.info("Register ZKstateListenter successfully!");
    }

    /**
     * 根据path获取值
     * @param path
     * @return
     * @throws Exception
     */
    public String read(String path) throws Exception {
        if (curator.checkExists().forPath(path) == null) {
            log.debug("path [{}] is not exists,return null", path);
            throw new IllegalArgumentException("Path:" + path + " content is empty, error!");
        } else {
            log.debug("read from zookeeper,path=" + path);
            byte[] b = curator.getData().forPath(path);
            String value = new String(b, "utf-8");
            return value;
        }
    }

    public List<String> getChildren(String path) throws Exception {
        //检查路径是否存在
        if (curator.checkExists().forPath(path) == null) {
            log.debug("path [{}] is not exists,return null", path);
            return null;
        } else {
            List<String> children = curator.getChildren().forPath(path);
            return children;
        }
    }

    public List<String> getDBconfig(String appName) throws Exception {
        List<String> result;
        result = getChildren(Constant.getAppDB(appName));
        if (result.size() == 0 || result == null)
            throw new IllegalStateException(appName + " Dbconfig is empty!");
        return result;
    }

    /**
     * 获取路径
     * @param paths
     * @return
     * @throws Exception
     */
    public String readWithPath(String... paths) throws Exception {
        if (paths.length == 0)
            return null;
        String path = null;
        for (String temp : paths) {
            if (path == null)
                path = Constant.PATH_SPLIT + temp;
            else
                path = path + Constant.PATH_SPLIT + temp;
        }
        path = path.replaceAll("//", "/");
        return read(path);
    }
}
